<?php

/**
 * AckHandler module.
 *
 * This file is part of MadelineProto.
 * MadelineProto is free software: you can redistribute it and/or modify it under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 * MadelineProto is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Affero General Public License for more details.
 * You should have received a copy of the GNU General Public License along with MadelineProto.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * @author    Daniil Gentili <daniil@daniil.it>
 * @copyright 2016-2020 Daniil Gentili <daniil@daniil.it>
 * @license   https://opensource.org/licenses/AGPL-3.0 AGPLv3
 *
 * @link https://docs.madelineproto.xyz MadelineProto documentation
 */

namespace danog\MadelineProto\MTProtoSession;

use danog\MadelineProto\DataCenterConnection;

/**
 * Manages acknowledgement of messages.
 *
 * @property DataCenterConnection $shared
 */
trait AckHandler
{
    /**
     * Acknowledge outgoing message ID.
     *
     * @param string|int $message_id Message Id
     *
     * @return boolean
     */
    public function ackOutgoingMessageId($message_id): bool
    {
        // The server acknowledges that it received my message
        if (!isset($this->outgoing_messages[$message_id])) {
            $this->logger->logger("WARNING: Couldn't find message id ".$message_id.' in the array of outgoing messages. Maybe try to increase its size?', \danog\MadelineProto\Logger::WARNING);
            return false;
        }
        return true;
    }
    /**
     * We have gotten response for outgoing message ID.
     *
     * @param string|int $message_id Message ID
     *
     * @return boolean
     */
    public function gotResponseForOutgoingMessageId($message_id): bool
    {
        // The server acknowledges that it received my message
        if (isset($this->new_outgoing[$message_id])) {
            unset($this->new_outgoing[$message_id]);
        }
        if (!isset($this->outgoing_messages[$message_id])) {
            $this->logger->logger("WARNING: Couldn't find message id ".$message_id.' in the array of outgoing messages. Maybe try to increase its size?', \danog\MadelineProto\Logger::WARNING);
            return false;
        }
        if (isset($this->outgoing_messages[$message_id]['body'])) {
            unset($this->outgoing_messages[$message_id]['body']);
        }
        if (isset($this->outgoing_messages[$message_id]['serialized_body'])) {
            unset($this->outgoing_messages[$message_id]['serialized_body']);
        }
        return true;
    }
    /**
     * Acknowledge incoming message ID.
     *
     * @param string|int $message_id Message ID
     *
     * @return boolean
     */
    public function ackIncomingMessageId($message_id): bool
    {
        // I let the server know that I received its message
        if (!isset($this->incoming_messages[$message_id])) {
            $this->logger->logger("WARNING: Couldn't find message id ".$message_id.' in the array of incoming messages. Maybe try to increase its size?', \danog\MadelineProto\Logger::WARNING);
        }
        $this->ack_queue[$message_id] = $message_id;
        return true;
    }

    /**
     * Check if there are some pending calls.
     *
     * @return boolean
     */
    public function hasPendingCalls(): bool
    {
        $timeout = $this->shared->getSettings()->getTimeout();
        $pfs = $this->shared->getGenericSettings()->getAuth()->getPfs();
        $unencrypted = !$this->shared->hasTempAuthKey();
        $notBound = !$this->shared->isBound();
        $pfsNotBound = $pfs && $notBound;
        foreach ($this->new_outgoing as $message_id) {
            if (isset($this->outgoing_messages[$message_id]['sent']) && $this->outgoing_messages[$message_id]['sent'] + $timeout < \time() && $unencrypted === $this->outgoing_messages[$message_id]['unencrypted'] && $this->outgoing_messages[$message_id]['_'] !== 'msgs_state_req') {
                if ($pfsNotBound && $this->outgoing_messages[$message_id]['_'] !== 'auth.bindTempAuthKey') {
                    continue;
                }
                return true;
            }
        }
        return false;
    }
    /**
     * Get all pending calls (also clear pending state requests).
     *
     * @return array
     */
    public function getPendingCalls(): array
    {
        $settings = $this->shared->getSettings();
        $global = $this->shared->getGenericSettings();
        $dropTimeout = $global->getRpc()->getRpcTimeout();
        $timeout = $settings->getTimeout();
        $pfs = $global->getAuth()->getPfs();
        $unencrypted = !$this->shared->hasTempAuthKey();
        $notBound = !$this->shared->isBound();
        $pfsNotBound = $pfs && $notBound;
        $result = [];
        foreach ($this->new_outgoing as $k => $message_id) {
            if (isset($this->outgoing_messages[$message_id]['sent']) && $this->outgoing_messages[$message_id]['sent'] + $timeout < \time() && $unencrypted === $this->outgoing_messages[$message_id]['unencrypted']) {
                if ($pfsNotBound && $this->outgoing_messages[$message_id]['_'] !== 'auth.bindTempAuthKey') {
                    continue;
                }
                if ($this->outgoing_messages[$message_id]['_'] === 'msgs_state_req') {
                    unset($this->new_outgoing[$k], $this->outgoing_messages[$message_id]);
                    continue;
                }
                if ($this->outgoing_messages[$message_id]['sent'] + $dropTimeout < \time()) {
                    $this->gotResponseForOutgoingMessageId($message_id);
                    $this->handleReject($this->outgoing_messages[$message_id], new \danog\MadelineProto\Exception("Request timeout"));
                    continue;
                }
                $result[] = $message_id;
            }
        }
        return $result;
    }
}
