<?php

/**
 * Session paths module.
 *
 * This file is part of MadelineProto.
 * MadelineProto is free software: you can redistribute it and/or modify it under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 * MadelineProto is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Affero General Public License for more details.
 * You should have received a copy of the GNU General Public License along with MadelineProto.
 * If not, see <http://www.gnu.org/licenses/>.
 *
 * @author    Daniil Gentili <daniil@daniil.it>
 * @copyright 2016-2020 Daniil Gentili <daniil@daniil.it>
 * @license   https://opensource.org/licenses/AGPL-3.0 AGPLv3
 *
 * @link https://docs.madelineproto.xyz MadelineProto documentation
 */

namespace danog\MadelineProto;

use Amp\Promise;
use danog\MadelineProto\Ipc\LightState;

/**
 * Session path information.
 */
class SessionPaths
{
    /**
     * Legacy session path.
     */
    private string $legacySessionPath;
    /**
     * Session path.
     */
    private string $sessionPath;
    /**
     * Session lock path.
     */
    private string $lockPath;
    /**
     * IPC socket path.
     */
    private string $ipcPath;
    /**
     * IPC light state path.
     */
    private string $ipcStatePath;
    /**
     * Temporary serialization path.
     */
    private string $tempPath;
    /**
     * Construct session info from session name.
     *
     * @param string $session Session name
     */
    public function __construct(string $session)
    {
        $session = Tools::absolute($session);
        $this->legacySessionPath = $session;
        $this->sessionPath = "$session.safe.php";
        $this->lockPath = "$session.lock";
        $this->ipcPath = "$session.ipc";
        $this->ipcStatePath = "$session.ipcState.php";
        $this->tempPath = "$session.temp.php";
    }
    /**
     * Get session path.
     *
     * @return string
     */
    public function __toString(): string
    {
        return $this->legacySessionPath;
    }

    /**
     * Get legacy session path.
     *
     * @return string
     */
    public function getLegacySessionPath(): string
    {
        return $this->legacySessionPath;
    }

    /**
     * Get session path.
     *
     * @return string
     */
    public function getSessionPath(): string
    {
        return $this->sessionPath;
    }

    /**
     * Get lock path.
     *
     * @return string
     */
    public function getLockPath(): string
    {
        return $this->lockPath;
    }

    /**
     * Get IPC socket path.
     *
     * @return string
     */
    public function getIpcPath(): string
    {
        return $this->ipcPath;
    }

    /**
     * Get temporary serialization path.
     *
     * @return string
     */
    public function getTempPath(): string
    {
        return $this->tempPath;
    }

    /**
     * Get IPC light state path.
     *
     * @return string
     */
    public function getIpcStatePath(): string
    {
        return $this->ipcStatePath;
    }

    /**
     * Get IPC state.
     *
     * @return Promise<LightState>
     */
    public function getIpcState(): Promise
    {
        return Tools::call(Serialization::newUnserialize($this->ipcStatePath));
    }
}
